@@ -1,9 +1,7 @@
 /* src/cacaoh/cacaoh.c - main for header generation (cacaoh)
 
-   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
-   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
-   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
-   J. Wenninger, Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006, 2007, 2008
+   CACAOVM - Verein zur Foerderung der freien virtuellen Maschine CACAO
 
    This file is part of CACAO.
 
@@ -22,8 +20,6 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: cacaoh.c 7246 2007-01-29 18:49:05Z twisti $
-
 */
 
 
@@ -55,12 +51,6 @@
 #include "vmcore/suck.h"
 
 
-/* define heap sizes **********************************************************/
-
-#define HEAP_MAXSIZE      2 * 1024 * 1024   /* default 2MB                    */
-#define HEAP_STARTSIZE         100 * 1024   /* default 100kB                  */
-
-
 /* define cacaoh options ******************************************************/
 
 enum {
@@ -163,8 +153,6 @@ int main(int argc, char **argv)
 	char *classpath;
 	char *cp;
 	s4    cplen;
-	u4    heapmaxsize;
-	u4    heapstartsize;
 
 	if (argc < 2)
 		usage();
@@ -179,19 +167,19 @@ int main(int argc, char **argv)
 	}
 	else {
 		cplen =
-#if defined(WITH_CLASSPATH_GNU)
+#if defined(WITH_JAVA_RUNTIME_LIBRARY_GNU_CLASSPATH)
 			strlen(CACAO_VM_ZIP) +
 			strlen(":") +
 #endif
-			strlen(CLASSPATH_CLASSES) +
+			strlen(JAVA_RUNTIME_LIBRARY_CLASSES) +
 			strlen("0");
 
 		bootclasspath = MNEW(char, cplen);
-#if defined(WITH_CLASSPATH_GNU)
+#if defined(WITH_JAVA_RUNTIME_LIBRARY_GNU_CLASSPATH)
 		strcat(bootclasspath, CACAO_VM_ZIP);
 		strcat(bootclasspath, ":");
 #endif
-		strcat(bootclasspath, CLASSPATH_CLASSES);
+		strcat(bootclasspath, JAVA_RUNTIME_LIBRARY_CLASSES);
 	}
 
 
@@ -214,9 +202,6 @@ int main(int argc, char **argv)
 	opt_verbose = false;
 	opt_directory = NULL;
 
-	heapmaxsize = HEAP_MAXSIZE;
-	heapstartsize = HEAP_STARTSIZE;
-
 
 	/* parse the options ******************************************************/
 
@@ -260,7 +245,6 @@ int main(int argc, char **argv)
 		case OPT_VERBOSE:
 			opt_verbose = true;
 			loadverbose = true;
-			linkverbose = true;
 			break;
 
 		default:
@@ -275,11 +259,7 @@ int main(int argc, char **argv)
 		log_println("Java - header-generator started"); 
 	}
 
-	/* initialize the utf8 hashtable stuff: lock, often used utf8 strings
-	   (must be done _after_ threads_preinit) */
-
-	if (!utf8_init())
-		vm_abort("utf8_init failed\n");
+	utf8_init();
 
 	/* initialize the classcache hashtable stuff: lock, hashtable
 	   (must be done _after_ threads_preinit) */
@@ -300,11 +280,10 @@ int main(int argc, char **argv)
 
 	suck_add(classpath);
 
-	/* initialize the loader subsystems (must be done _after_
-       classcache_init) */
+	/* AFTER: classcache_init */
 
-	if (!loader_init())
-		vm_abort("loader_init failed\n");
+	loader_preinit();
+	loader_init();
 
 
 	/* load Java classes ******************************************************/
