@@ -1,9 +1,7 @@
 /* src/cacaoh/headers.c - functions for header generation
 
-   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
-   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
-   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
-   J. Wenninger, Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006, 2007, 2008
+   CACAOVM - Verein zur Foerderung der freien virtuellen Maschine CACAO
 
    This file is part of CACAO.
 
@@ -22,8 +20,6 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: headers.c 7462 2007-03-05 23:56:41Z edwin $
-
 */
 
 
@@ -32,11 +28,10 @@
 #include <assert.h>
 #include <ctype.h>
 #include <stdarg.h>
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "vm/types.h"
-
 #if defined(ENABLE_THREADS)
 # if defined(__DARWIN__)
 #  include <signal.h>
@@ -64,9 +59,15 @@
 
 #define ACC_NATIVELY_OVERLOADED    0x10000000
 
+#if defined(ENABLE_HANDLES)
+# define HEAP_PREFIX "heap_"
+#else
+# define HEAP_PREFIX ""
+#endif
+
 chain *ident_chain;     /* chain with method and field names in current class */
 FILE *file = NULL;
-static u4 outputsize;
+static uint32_t outputsize;
 static bool dopadding;
 
 
@@ -92,14 +93,19 @@ void printID(utf *u)
 }
 
 
-static void addoutputsize (int len)
+static void addoutputsize(int len)
 {
-	u4 newsize,i;
-	if (!dopadding) return;
+	uint32_t newsize;
+	int32_t  i;
+
+	if (!dopadding)
+		return;
 
 	newsize = MEMORY_ALIGN(outputsize, len);
 	
-	for (i = outputsize; i < newsize; i++) fprintf(file, "   u1 pad%d\n", (int) i);
+	for (i = outputsize; i < newsize; i++)
+		fprintf(file, "   uint8_t pad%d\n", (int) i);
+
 	outputsize = newsize;
 }
 
@@ -107,7 +113,7 @@ static void addoutputsize (int len)
 void printOverloadPart(utf *desc)
 {
 	char *utf_ptr=desc->text;
-	u2 c;
+	uint16_t c;
 
 	fprintf(file, "__");
 
@@ -141,48 +147,53 @@ void printOverloadPart(utf *desc)
 	}
 }
 
-static char *printtype(char *utf_ptr)
+static char *printtype(char *utf_ptr, char *prefix, char *infix)
 {
-	u2 c;
+	uint16_t c;
 
 	switch (utf_nextu2(&utf_ptr)) {
-	case 'V': fprintf (file, "void");
+	case 'V':
+		fprintf(file, "void");
 		break;
 	case 'I':
 	case 'S':
 	case 'B':
 	case 'C':
-	case 'Z': addoutputsize (4);
-		fprintf (file, "s4");
+	case 'Z':
+		addoutputsize(4);
+		fprintf(file, "int32_t");
 		break;
-	case 'J': addoutputsize (8);
-		fprintf (file, "s8");
+	case 'J':
+		addoutputsize(8);
+		fprintf(file, "int64_t");
 		break;
-	case 'F': addoutputsize (4);
-		fprintf (file, "float");
+	case 'F':
+		addoutputsize(4);
+		fprintf(file, "float");
 		break;
-	case 'D': addoutputsize (8);
-		fprintf (file, "double");
+	case 'D':
+		addoutputsize(8);
+		fprintf(file, "double");
 		break;
 	case '[':
-		addoutputsize ( sizeof(java_arrayheader*) ); 
+		addoutputsize ( sizeof(java_array_t*) ); 
 		switch (utf_nextu2(&utf_ptr)) {
-		case 'I':  fprintf (file, "java_intarray*"); break;
-		case 'J':  fprintf (file, "java_longarray*"); break;
-		case 'Z':  fprintf (file, "java_booleanarray*"); break;
-		case 'B':  fprintf (file, "java_bytearray*"); break;
-		case 'S':  fprintf (file, "java_shortarray*"); break;
-		case 'C':  fprintf (file, "java_chararray*"); break;
-		case 'F':  fprintf (file, "java_floatarray*"); break;
-		case 'D':  fprintf (file, "java_doublearray*"); break;
+		case 'I':  fprintf (file, "java%s_intarray_t*", infix); break;
+		case 'J':  fprintf (file, "java%s_longarray_t*", infix); break;
+		case 'Z':  fprintf (file, "java%s_booleanarray_t*", infix); break;
+		case 'B':  fprintf (file, "java%s_bytearray_t*", infix); break;
+		case 'S':  fprintf (file, "java%s_shortarray_t*", infix); break;
+		case 'C':  fprintf (file, "java%s_chararray_t*", infix); break;
+		case 'F':  fprintf (file, "java%s_floatarray_t*", infix); break;
+		case 'D':  fprintf (file, "java%s_doublearray_t*", infix); break;
 				
-		case '[': fprintf(file, "java_objectarray*");
+		case '[': fprintf(file, "java%s_objectarray_t*", infix);
 			while ((c = utf_nextu2(&utf_ptr)) == '[');
 			if (c == 'L')
 				while (utf_nextu2(&utf_ptr) != ';');
 			break;
                            
-		case 'L':  fprintf(file, "java_objectarray*");
+		case 'L':  fprintf(file, "java%s_objectarray_t*", infix);
 			while (utf_nextu2(&utf_ptr) != ';');
 			break;
 		default:
@@ -192,8 +203,8 @@ static char *printtype(char *utf_ptr)
 		break;
 		
 	case 'L': 
-		addoutputsize ( sizeof(java_objectheader*));
-		fprintf (file, "struct ");
+		addoutputsize ( sizeof(java_object_t*));
+		fprintf (file, "struct %s", prefix);
 		while ( (c = utf_nextu2(&utf_ptr)) != ';' ) printIDpart (c);   	 
 		fprintf (file, "*");
 		break;
@@ -227,24 +238,24 @@ static int searchidentchain_utf(utf *ident)
 
 static void printfields(classinfo *c)
 {
-	u4 i;
+	int32_t i;
 	fieldinfo *f;
 	int ident_count;
 	
 	if (!c) {
-		addoutputsize(sizeof(java_objectheader));
-		fprintf(file, "   java_objectheader header;\n");
+		addoutputsize(sizeof(java_object_t));
+		fprintf(file, "   java_object_t header;\n");
 		return;
 	}
 		
-	printfields(c->super.cls);
+	printfields(c->super);
 	
 	for (i = 0; i < c->fieldscount; i++) {
 		f = &(c->fields[i]);
 		
 		if (!(f->flags & ACC_STATIC)) {
 			fprintf(file, "   ");
-			printtype(f->descriptor->text);
+			printtype(f->descriptor->text, HEAP_PREFIX, "");
 			fprintf(file, " ");
 			utf_fprint_printable_ascii(file, f->name);
 
@@ -264,7 +275,7 @@ static void printfields(classinfo *c)
 void printmethod(methodinfo *m)
 {
 	char *utf_ptr;
-	u2 paramnum = 1;
+	int32_t paramnum = 1;
 
 	/* search for return-type in descriptor */	
 	utf_ptr = m->descriptor->text;
@@ -272,7 +283,7 @@ void printmethod(methodinfo *m)
 
 	/* create remarks */
 	fprintf(file, "\n/*\n * Class:     ");
-	utf_fprint_printable_ascii(file, m->class->name);
+	utf_fprint_printable_ascii(file, m->clazz->name);
 	fprintf(file, "\n * Method:    ");
 	utf_fprint_printable_ascii(file, m->name);
 	fprintf(file, "\n * Signature: ");
@@ -281,9 +292,9 @@ void printmethod(methodinfo *m)
 
 	/* create prototype */ 			
 	fprintf(file, "JNIEXPORT ");
-	printtype(utf_ptr);
+	printtype(utf_ptr, "", "_handle");
 	fprintf(file, " JNICALL Java_");
-	printID(m->class->name);
+	printID(m->clazz->name);
 
 	chain_addlast(ident_chain, m->name);
 
@@ -301,7 +312,7 @@ void printmethod(methodinfo *m)
 			
 	if (!(m->flags & ACC_STATIC)) {
 		fprintf(file, ", struct ");
-		printID(m->class->name);
+		printID(m->clazz->name);
 		fprintf(file, "* this");
 
 	} else {
@@ -311,7 +322,7 @@ void printmethod(methodinfo *m)
 	if ((*utf_ptr) != ')') fprintf(file, ", ");
 			
 	while ((*utf_ptr) != ')') {
-		utf_ptr = printtype(utf_ptr);
+		utf_ptr = printtype(utf_ptr, "", "_handle");
 		fprintf(file, " par%d", paramnum++);
 		if ((*utf_ptr)!=')') fprintf(file, ", ");
 	}
@@ -324,7 +335,7 @@ void printmethod(methodinfo *m)
 
 void gen_header_filename(char *buffer, utf *u)
 {
-	s4 i;
+	int32_t i;
   
 	for (i = 0; i < utf_get_number_of_u2s(u); i++) {
 		if ((u->text[i] == '/') || (u->text[i] == '$')) {
@@ -345,9 +356,9 @@ void headerfile_generate(classinfo *c, char *opt_directory)
 	char header_filename[1024] = "";
 	char classname[1024]; 
 	char uclassname[1024];
-	u2 i;
+	int32_t i;
 	methodinfo *m;	      		
-	u2 j;
+	int32_t j;
 	methodinfo *m2;
 	bool nativelyoverloaded;
 
@@ -388,17 +399,30 @@ void headerfile_generate(classinfo *c, char *opt_directory)
 	fprintf(file, "/* Structure information for class: ");
 	utf_fprint_printable_ascii(file, c->name);
 	fprintf(file, " */\n\n");
-	fprintf(file, "typedef struct ");
-	printID(c->name);							
+	fprintf(file, "typedef struct %s", HEAP_PREFIX);
+	printID(c->name);
 	fprintf(file, " {\n");
 	outputsize = 0;
 	dopadding = true;
 
 	printfields(c);
 
+	fprintf(file, "} %s", HEAP_PREFIX);
+	printID(c->name);
+	fprintf(file, ";\n\n");
+
+#if defined(ENABLE_HANDLES)
+	/* create structure for indirection cell */
+	fprintf(file, "typedef struct ");
+	printID(c->name);
+	fprintf(file, " {\n");
+	fprintf(file, "   %s", HEAP_PREFIX);
+	printID(c->name);
+	fprintf(file, " *heap_object;\n");
 	fprintf(file, "} ");
 	printID(c->name);
 	fprintf(file, ";\n\n");
+#endif
 
 	/* create chain for renaming overloaded methods */
 	chain_free(ident_chain);
@@ -458,7 +482,7 @@ void print_classname(classinfo *clazz)
 	utf *u = clazz->name;
     char *endpos  = u->text + u->blength;
     char *utf_ptr = u->text; 
-	u2 c;
+	uint16_t c;
 
     while (utf_ptr < endpos) {
 		if ((c = utf_nextu2(&utf_ptr)) == '_')
