@@ -22,14 +22,14 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: arch.h 7695 2007-04-12 19:49:34Z twisti $
-
 */
 
 
 #ifndef _ARCH_H
 #define _ARCH_H
 
+#define JIT_COMPILER_VIA_SIGNAL
+
 #include "config.h"
 
 
@@ -104,12 +104,6 @@
 #endif
 
 
-#define HAS_4BYTE_STACKSLOT
-#define ALIGN_LONGS_IN_MEMORY       /* Align Longs and/or Doubles at         */
-#define ALIGN_DOUBLES_IN_MEMORY     /* 2*Stackslotsize relativ to stackframe */
-/* Memory Positions for not Interface Stackslots (allocate_scratch_registers)*/
-/* are not properly aligned in case HAS_4_BYTE_STACKSLOT is not defined!     */
-/* For HAS_4_BYTE_STACKSLOT archs no distinction is made between long and dbl*/
 /* define SUPPORT_COMBINE_INTEGER_REGISTERS */
 
 
@@ -128,8 +122,11 @@
 
 /* replacement ****************************************************************/
 
-#define REPLACEMENT_PATCH_SIZE           5       /* bytes */
-#define REPLACEMENT_STUB_SIZE            (5+5+1) /* bytes */
+#define REPLACEMENT_PATCH_SIZE           2       /* bytes */
+
+/* subtype ********************************************************************/
+
+#define USES_NEW_SUBTYPE                 1
 
 #endif /* _ARCH_H */
 
