@@ -1,9 +1,7 @@
 /* src/vm/jit/i386/darwin/md-asm.h - assembler defines for i386 Darwin ABI
 
-   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
-   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
-   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
-   J. Wenninger, Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006, 2008
+   CACAOVM - Verein zur Foerderung der freien virtuellen Maschine CACAO
 
    This file is part of CACAO.
 
@@ -22,14 +20,6 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   Contact: cacao@cacaojvm.org
-
-   Authors: Christian Thalinger
-
-   Changes:
-
-   $Id: md-asm.h 4357 2006-01-22 23:33:38Z twisti $
-
 */
 
 
@@ -98,24 +88,18 @@
 
 #define asm_patcher_wrapper                   _asm_patcher_wrapper
 
-#define asm_replacement_out                   _asm_replacement_out
-#define asm_replacement_in                    _asm_replacement_in
-
 #define asm_builtin_f2i                       _asm_builtin_f2i
 #define asm_builtin_f2l                       _asm_builtin_f2l
 #define asm_builtin_d2i                       _asm_builtin_d2i
 #define asm_builtin_d2l                       _asm_builtin_d2l
 
-#define asm_criticalsections                  _asm_criticalsections
-#define asm_getclassvalues_atomic             _asm_getclassvalues_atomic
-
 
 /* external defines ***********************************************************/
 
 #define exceptions_get_and_clear_exception    _exceptions_get_and_clear_exception
 
 #define builtin_throw_exception               _builtin_throw_exception
-#define codegen_get_pv_from_pc                _codegen_get_pv_from_pc
+#define methodtree_find                       _methodtree_find
 #define exceptions_handle_exception           _exceptions_handle_exception
 #define jit_asm_compile                       _jit_asm_compile
 
@@ -133,6 +117,8 @@
 #define builtin_d2i                           _builtin_d2i
 #define builtin_d2l                           _builtin_d2l
 
+#define asm_get_cycle_count                   _asm_get_cycle_count
+
 #endif /* _MD_ASM_H */
 
 
