@@ -1,9 +1,7 @@
 /* src/vm/jit/i386/linux/md-os.c - machine dependent i386 Linux functions
 
-   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
-   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
-   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
-   J. Wenninger, Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006, 2007, 2008
+   CACAOVM - Verein zur Foerderung der freien virtuellen Maschine CACAO
 
    This file is part of CACAO.
 
@@ -22,8 +20,6 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: md-os.c 7964 2007-05-24 10:23:42Z twisti $
-
 */
 
 
@@ -31,20 +27,24 @@
 
 #include "config.h"
 
+#include <stdint.h>
 #include <ucontext.h>
 
 #include "vm/types.h"
 
 #include "vm/jit/i386/codegen.h"
+#include "vm/jit/i386/md.h"
 
-#include "threads/threads-common.h"
+#include "threads/thread.h"
 
-#include "vm/exceptions.h"
+#include "vm/builtin.h"
 #include "vm/signallocal.h"
 #include "vm/stringlocal.h"
 
 #include "vm/jit/asmpart.h"
+#include "vm/jit/executionstate.h"
 #include "vm/jit/stacktrace.h"
+#include "vm/jit/trap.h"
 
 
 /* md_signal_handler_sigsegv ***************************************************
@@ -55,20 +55,21 @@
 
 void md_signal_handler_sigsegv(int sig, siginfo_t *siginfo, void *_p)
 {
-	ucontext_t        *_uc;
-	mcontext_t        *_mc;
-	u1                *pv;
-	u1                *sp;
-	u1                *ra;
-	u1                *xpc;
-	u1                 opc;
-	u1                 mod;
-	u1                 rm;
-	s4                 d;
-	s4                 disp;
-	ptrint             val;
-	s4                 type;
-	java_objectheader *o;
+	ucontext_t     *_uc;
+	mcontext_t     *_mc;
+	u1             *pv;
+	u1             *sp;
+	u1             *ra;
+	u1             *xpc;
+	u1              opc;
+	u1              mod;
+	u1              rm;
+	s4              d;
+	s4              disp;
+	ptrint          val;
+	s4              type;
+	void           *p;
+	java_object_t  *o;
 
 	_uc = (ucontext_t *) _p;
 	_mc = &_uc->uc_mcontext;
@@ -102,23 +103,59 @@ void md_signal_handler_sigsegv(int sig, siginfo_t *siginfo, void *_p)
 		   here. */
 
 		val = _mc->gregs[REG_EAX - d];
+
+		if (type == TRAP_COMPILER) {
+			/* The PV from the compiler stub is equal to the XPC. */
+
+			pv = xpc;
+
+			/* We use a framesize of zero here because the call pushed
+			   the return addres onto the stack. */
+
+			ra = md_stacktrace_get_returnaddress(sp, 0);
+
+			/* Skip the RA on the stack. */
+
+			sp = sp + 1 * SIZEOF_VOID_P;
+
+			/* The XPC is the RA minus 2, because the RA points to the
+			   instruction after the call. */
+
+			xpc = ra - 2;
+		}
 	}
 	else {
 		/* this was a normal NPE */
 
-		type = EXCEPTION_HARDWARE_NULLPOINTER;
+		type = TRAP_NullPointerException;
 		val  = 0;
 	}
 
-	/* generate appropriate exception */
+	/* Handle the trap. */
 
-	o = exceptions_new_hardware_exception(pv, sp, ra, xpc, type, val);
+	p = trap_handle(type, val, pv, sp, ra, xpc, _p);
 
-	/* set registers */
+	/* Set registers. */
 
-	_mc->gregs[REG_EAX] = (ptrint) o;
-	_mc->gregs[REG_ECX] = (ptrint) xpc;                      /* REG_ITMP2_XPC */
-	_mc->gregs[REG_EIP] = (ptrint) asm_handle_exception;
+	if (type == TRAP_COMPILER) {
+		if (p == NULL) {
+			o = builtin_retrieve_exception();
+
+			_mc->gregs[REG_ESP] = (uintptr_t) sp;    /* Remove RA from stack. */
+
+			_mc->gregs[REG_EAX] = (uintptr_t) o;
+			_mc->gregs[REG_ECX] = (uintptr_t) xpc;           /* REG_ITMP2_XPC */
+			_mc->gregs[REG_EIP] = (uintptr_t) asm_handle_exception;
+		}
+		else {
+			_mc->gregs[REG_EIP] = (uintptr_t) p;
+		}
+	}
+	else {
+		_mc->gregs[REG_EAX] = (uintptr_t) p;
+		_mc->gregs[REG_ECX] = (uintptr_t) xpc;               /* REG_ITMP2_XPC */
+		_mc->gregs[REG_EIP] = (uintptr_t) asm_handle_exception;
+	}
 }
 
 
@@ -131,15 +168,15 @@ void md_signal_handler_sigsegv(int sig, siginfo_t *siginfo, void *_p)
 
 void md_signal_handler_sigfpe(int sig, siginfo_t *siginfo, void *_p)
 {
-	ucontext_t        *_uc;
-	mcontext_t        *_mc;
-	u1                *pv;
-	u1                *sp;
-	u1                *ra;
-	u1                *xpc;
-	s4                 type;
-	ptrint             val;
-	java_objectheader *o;
+	ucontext_t     *_uc;
+	mcontext_t     *_mc;
+	u1             *pv;
+	u1             *sp;
+	u1             *ra;
+	u1             *xpc;
+	s4              type;
+	ptrint          val;
+	void           *p;
 
 	_uc = (ucontext_t *) _p;
 	_mc = &_uc->uc_mcontext;
@@ -149,21 +186,93 @@ void md_signal_handler_sigfpe(int sig, siginfo_t *siginfo, void *_p)
 	xpc = (u1 *) _mc->gregs[REG_EIP];
 	ra  = xpc;                          /* return address is equal to xpc     */
 
-	/* this is an ArithmeticException */
+	/* This is an ArithmeticException. */
 
-	type = EXCEPTION_HARDWARE_ARITHMETIC;
+	type = TRAP_ArithmeticException;
 	val  = 0;
 
-	/* generate appropriate exception */
+	/* Handle the trap. */
 
-	o = exceptions_new_hardware_exception(pv, sp, ra, xpc, type, val);
+	p = trap_handle(type, val, pv, sp, ra, xpc, _p);
 
-	_mc->gregs[REG_EAX] = (ptrint) o;
-	_mc->gregs[REG_ECX] = (ptrint) xpc;                      /* REG_ITMP2_XPC */
-	_mc->gregs[REG_EIP] = (ptrint) asm_handle_exception;
+	/* Set registers. */
+
+	_mc->gregs[REG_EAX] = (uintptr_t) p;
+	_mc->gregs[REG_ECX] = (uintptr_t) xpc;                   /* REG_ITMP2_XPC */
+	_mc->gregs[REG_EIP] = (uintptr_t) asm_handle_exception;
 }
 
 
+/* md_signal_handler_sigill ****************************************************
+
+   Signal handler for hardware patcher traps (ud2).
+
+*******************************************************************************/
+
+void md_signal_handler_sigill(int sig, siginfo_t *siginfo, void *_p)
+{
+	ucontext_t        *_uc;
+	mcontext_t        *_mc;
+	u1                *pv;
+	u1                *sp;
+	u1                *ra;
+	u1                *xpc;
+	s4                 type;
+	ptrint             val;
+	void              *p;
+
+	_uc = (ucontext_t *) _p;
+	_mc = &_uc->uc_mcontext;
+
+	pv  = NULL;                 /* is resolved during stackframeinfo creation */
+	sp  = (u1 *) _mc->gregs[REG_ESP];
+	xpc = (u1 *) _mc->gregs[REG_EIP];
+	ra  = xpc;                            /* return address is equal to xpc   */
+
+	type = TRAP_PATCHER;
+	val  = 0;
+
+	/* Handle the trap. */
+
+	p = trap_handle(type, val, pv, sp, ra, xpc, _p);
+
+	/* Set registers. */
+
+	if (p != NULL) {
+		_mc->gregs[REG_EAX] = (uintptr_t) p;
+		_mc->gregs[REG_ECX] = (uintptr_t) xpc;               /* REG_ITMP2_XPC */
+		_mc->gregs[REG_EIP] = (uintptr_t) asm_handle_exception;
+	}
+}
+
+
+/* md_signal_handler_sigusr1 ***************************************************
+
+   Signal handler for suspending threads.
+
+*******************************************************************************/
+
+#if defined(ENABLE_THREADS) && defined(ENABLE_GC_CACAO)
+void md_signal_handler_sigusr1(int sig, siginfo_t *siginfo, void *_p)
+{
+	ucontext_t *_uc;
+	mcontext_t *_mc;
+	u1         *pc;
+	u1         *sp;
+
+	_uc = (ucontext_t *) _p;
+	_mc = &_uc->uc_mcontext;
+
+	/* get the PC and SP for this thread */
+	pc = (u1 *) _mc->gregs[REG_EIP];
+	sp = (u1 *) _mc->gregs[REG_ESP];
+
+	/* now suspend the current thread */
+	threads_suspend_ack(pc, sp);
+}
+#endif
+
+
 /* md_signal_handler_sigusr2 ***************************************************
 
    Signal handler for profiling sampling.
@@ -190,30 +299,59 @@ void md_signal_handler_sigusr2(int sig, siginfo_t *siginfo, void *_p)
 #endif
 
 
-/* md_critical_section_restart *************************************************
+/* md_executionstate_read ******************************************************
 
-   Search the critical sections tree for a matching section and set
-   the PC to the restart point, if necessary.
+   Read the given context into an executionstate for Replacement.
 
 *******************************************************************************/
 
-#if defined(ENABLE_THREADS)
-void md_critical_section_restart(ucontext_t *_uc)
+void md_executionstate_read(executionstate_t *es, void *context)
 {
+	ucontext_t *_uc;
 	mcontext_t *_mc;
-	u1         *pc;
-	u1         *npc;
+	s4          i;
 
+	_uc = (ucontext_t *) context;
 	_mc = &_uc->uc_mcontext;
 
-	pc = (u1 *) _mc->gregs[REG_EIP];
+	/* read special registers */
+	es->pc = (u1 *) _mc->gregs[REG_EIP];
+	es->sp = (u1 *) _mc->gregs[REG_ESP];
+	es->pv = NULL;                   /* pv must be looked up via AVL tree */
 
-	npc = critical_find_restart_point(pc);
+	/* read integer registers */
+	for (i = 0; i < INT_REG_CNT; i++)
+		es->intregs[i] = _mc->gregs[REG_EAX - i];
 
-	if (npc != NULL)
-		_mc->gregs[REG_EIP] = (ptrint) npc;
+	/* read float registers */
+	for (i = 0; i < FLT_REG_CNT; i++)
+		es->fltregs[i] = 0xdeadbeefdeadbeefULL;
+}
+
+
+/* md_executionstate_write *****************************************************
+
+   Write the given executionstate back to the context for Replacement.
+
+*******************************************************************************/
+
+void md_executionstate_write(executionstate_t *es, void *context)
+{
+	ucontext_t *_uc;
+	mcontext_t *_mc;
+	s4          i;
+
+	_uc = (ucontext_t *) context;
+	_mc = &_uc->uc_mcontext;
+
+	/* write integer registers */
+	for (i = 0; i < INT_REG_CNT; i++)
+		_mc->gregs[REG_EAX - i] = es->intregs[i];
+
+	/* write special registers */
+	_mc->gregs[REG_EIP] = (ptrint) es->pc;
+	_mc->gregs[REG_ESP] = (ptrint) es->sp;
 }
-#endif
 
 
 /*
