@@ -22,27 +22,21 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: md.c 7596 2007-03-28 21:05:53Z twisti $
-
 */
 
 
 #include "config.h"
 
 #include <assert.h>
+#include <stdint.h>
 
 #include "vm/types.h"
 
 #include "vm/global.h"
+#include "vm/vm.h"
 
 #include "vm/jit/asmpart.h"
-#include "vm/jit/codegen-common.h"
-#include "vm/jit/md.h"
-
-#if !defined(NDEBUG) && defined(ENABLE_DISASSEMBLER)
-#include "vmcore/options.h" /* XXX debug */
-#include "vm/jit/disass.h" /* XXX debug */
-#endif
+#include "vm/jit/jit.h"
 
 
 /* md_init *********************************************************************
@@ -57,26 +51,7 @@ void md_init(void)
 }
 
 
-/* md_stacktrace_get_returnaddress *********************************************
-
-   Returns the return address of the current stackframe, specified by
-   the passed stack pointer and the stack frame size.
-
-*******************************************************************************/
-
-u1 *md_stacktrace_get_returnaddress(u1 *sp, u4 framesize)
-{
-	u1 *ra;
-
-	/* on i386 the return address is above the current stack frame */
-
-	ra = *((u1 **) (sp + framesize));
-
-	return ra;
-}
-
-
-/* md_get_method_patch_address *************************************************
+/* md_jit_method_patch_address *************************************************
 
    Gets the patch address of the currently compiled method. The offset
    is extracted from the load instruction(s) before the jump and added
@@ -102,115 +77,60 @@ u1 *md_stacktrace_get_returnaddress(u1 *sp, u4 framesize)
 
 *******************************************************************************/
 
-u1 *md_get_method_patch_address(u1 *ra, stackframeinfo *sfi, u1 *mptr)
+void *md_jit_method_patch_address(void *pv, void *ra, void *mptr)
 {
-	u1  mcode;
-	s4  offset;
-	u1 *pa;                             /* patch address                      */
+	uint8_t  *pc;
+	uint16_t  opcode;
+	int32_t   disp;
+	void     *pa;                                            /* patch address */
 
 	/* go back to the actual call instruction (2-bytes) */
 
-	ra = ra - 2;
+	pc = ((uint8_t *) ra) - 2;
 
-	/* get the last byte of the call */
+	/* Get the opcode of the call. */
 
-	mcode = ra[1];
+	opcode = *((uint16_t *) pc);
 
 	/* check for the different calls */
 
-	if (mcode == 0xd1) {
+	switch (opcode) {
+	case 0xd1ff:
 		/* INVOKESTATIC/SPECIAL */
 
-		/* patch address is 4-bytes before the call instruction */
+		/* Patch address is 4-bytes before the call instruction. */
 
-		pa = ra - 4;
-	}
-	else if (mcode == 0xd2) {
+		pa = pc - 4;
+		break;
+
+	case 0xd2ff:
 		/* INVOKEVIRTUAL/INTERFACE */
 
-		/* return NULL if no mptr was specified (used for replacement) */
+		/* Return NULL if no mptr was specified (used for
+		   replacement). */
 
 		if (mptr == NULL)
 			return NULL;
 
-		/* Get the offset from the instruction (the offset address is
-		   4-bytes before the call instruction). */
+		/* Get the displacement from the instruction (the displacement
+		   address is 4-bytes before the call instruction). */
 
-		offset = *((s4 *) (ra - 4));
+		disp = *((int32_t *) (pc - 4));
 
-		/* add the offset to the method pointer */
+		/* Add the displacement to the method pointer. */
 
-		pa = mptr + offset;
-	}
-	else {
-		/* catch any problems */
+		pa = ((uint8_t *) mptr) + disp;
+		break;
 
-		pa = NULL; /* avoid warnings */
-
-		vm_abort("couldn't find a proper call instruction sequence");
+	default:
+		vm_abort_disassemble(pc, 1, "md_jit_method_patch_address: unknown instruction %x", opcode);
+		return NULL;
 	}
 
 	return pa;
 }
 
 
-/* md_codegen_get_pv_from_pc ***************************************************
-
-   On this architecture just a wrapper function to
-   codegen_get_pv_from_pc.
-
-*******************************************************************************/
-
-u1 *md_codegen_get_pv_from_pc(u1 *ra)
-{
-	u1 *pv;
-
-	/* Get the start address of the function which contains this
-       address from the method table. */
-
-	pv = codegen_get_pv_from_pc(ra);
-
-	return pv;
-}
-
-
-/* md_cacheflush ***************************************************************
-
-   Calls the system's function to flush the instruction and data
-   cache.
-
-*******************************************************************************/
-
-void md_cacheflush(u1 *addr, s4 nbytes)
-{
-	/* do nothing */
-}
-
-
-/* md_icacheflush **************************************************************
-
-   Calls the system's function to flush the instruction cache.
-
-*******************************************************************************/
-
-void md_icacheflush(u1 *addr, s4 nbytes)
-{
-	/* do nothing */
-}
-
-
-/* md_dcacheflush **************************************************************
-
-   Calls the system's function to flush the data cache.
-
-*******************************************************************************/
-
-void md_dcacheflush(u1 *addr, s4 nbytes)
-{
-	/* do nothing */
-}
-
-
 /* md_patch_replacement_point **************************************************
 
    Patch the given replacement point.
@@ -218,54 +138,26 @@ void md_dcacheflush(u1 *addr, s4 nbytes)
 *******************************************************************************/
 
 #if defined(ENABLE_REPLACEMENT)
-void md_patch_replacement_point(codeinfo *code, s4 index, rplpoint *rp, u1 *savedmcode)
+void md_patch_replacement_point(u1 *pc, u1 *savedmcode, bool revert)
 {
-	s4 disp;
-	u8 mcode;
+	u2 mcode;
 
-	/* XXX this is probably unsafe! */
-
-	if (index < 0) {
-		/* write spinning instruction */
-		*(u2*)(rp->pc) = 0xebfe;
-
-		/* write 5th byte */
-		rp->pc[4] = savedmcode[4];
-
-		/* write first word */
-		*(u4*)(rp->pc) = *(u4*)(savedmcode);
+	if (revert) {
+		/* write saved machine code */
+		*(u2*)(pc) = *(u2*)(savedmcode);
 	}
 	else {
 		/* save the current machine code */
-		*(u4*)(savedmcode) = *(u4*)(rp->pc);
-		savedmcode[4] = rp->pc[4];
+		*(u2*)(savedmcode) = *(u2*)(pc);
 
 		/* build the machine code for the patch */
-		disp = (code->replacementstubs - rp->pc)
-			   + index * REPLACEMENT_STUB_SIZE
-			   - 5;
+		mcode = 0x0b0f;
 
-		mcode = 0xe9 | ((u8) disp << 8);
-
-		/* write spinning instruction */
-		*(u2*)(rp->pc) = 0xebfe;
-
-		/* write 5th byte */
-		rp->pc[4] = (mcode >> 32);
-
-		/* write first word */
-		*(u4*)(rp->pc) = (u4) mcode;
+		/* write new machine code */
+		*(u2*)(pc) = mcode;
 	}
 
-#if !defined(NDEBUG) && defined(ENABLE_DISASSEMBLER) && 0
-	{
-		u1* u1ptr = rp->pc;
-		DISASSINSTR(u1ptr);
-		fflush(stdout);
-	}
-#endif
-			
-    /* XXX if required asm_cacheflush(rp->pc,8); */
+    /* XXX if required asm_cacheflush(pc,8); */
 }
 #endif /* defined(ENABLE_REPLACEMENT) */
 
