@@ -1,9 +1,7 @@
 /* src/vm/jit/mips/md-abi.c - functions for MIPS ABI
 
-   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
-   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
-   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
-   J. Wenninger, Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006, 2007, 2008
+   CACAOVM - Verein zur Foerderung der freien virtuellen Maschine CACAO
 
    This file is part of CACAO.
 
@@ -22,21 +20,27 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: md-abi.c 8013 2007-06-05 10:19:09Z twisti $
-
 */
 
 
 #include "config.h"
+
+#include <stdarg.h>
+#include <stdint.h>
+
 #include "vm/types.h"
 
 #include "vm/jit/mips/md-abi.h"
 
+#include "mm/memory.h"
+
 #include "vm/global.h"
 
 #include "vm/jit/abi.h"
+#include "vm/jit/stack.h"
 
 #include "vmcore/descriptor.h"
+#include "vmcore/method.h"
 
 
 /* register descripton array **************************************************/
@@ -289,7 +293,7 @@ void md_param_alloc(methoddesc *md)
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->regoff   = stacksize * 8;
 				stacksize++;
 			}
 			break;
@@ -304,7 +308,7 @@ void md_param_alloc(methoddesc *md)
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->regoff   = stacksize * 8;
 				stacksize++;
 			}
 			break;
@@ -331,13 +335,15 @@ void md_param_alloc(methoddesc *md)
 		case TYPE_ADR:
 			if (reguse < INT_ARG_CNT) {
 				pd->inmemory = false;
+				pd->index    = reguse;
 				pd->regoff   = abi_registers_integer_argument[reguse];
 				reguse++;
 				md->argintreguse = reguse;
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->index    = stacksize;
+				pd->regoff   = stacksize * 8;
 				stacksize++;
 			}
 			break;
@@ -348,10 +354,12 @@ void md_param_alloc(methoddesc *md)
 			if (reguse < INT_ARG_CNT) {
 				pd->inmemory = false;
 #  if WORDS_BIGENDIAN == 1
+				pd->index    = PACK_REGS(reguse + 1, reguse);
 				pd->regoff   =
 					PACK_REGS(abi_registers_integer_argument[reguse + 1],
 							  abi_registers_integer_argument[reguse]);
 #  else
+				pd->index    = PACK_REGS(reguse, reguse + 1);
 				pd->regoff   =
 					PACK_REGS(abi_registers_integer_argument[reguse],
 							  abi_registers_integer_argument[reguse + 1]);
@@ -361,7 +369,8 @@ void md_param_alloc(methoddesc *md)
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->index    = stacksize;
+				pd->regoff   = stacksize * 8;
 				stacksize++;
 			}
 			break;
@@ -370,13 +379,15 @@ void md_param_alloc(methoddesc *md)
 		case TYPE_DBL:
 			if (reguse < FLT_ARG_CNT) {
 				pd->inmemory = false;
+				pd->index    = reguse;
 				pd->regoff   = abi_registers_float_argument[reguse];
 				reguse++;
 				md->argfltreguse = reguse;
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->index    = stacksize;
+				pd->regoff   = stacksize * 8;
 				stacksize++;
 			}
 			break;
@@ -405,7 +416,7 @@ void md_param_alloc(methoddesc *md)
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->regoff   = stacksize * 8;
 			}
 			stacksize++;
 			break;
@@ -430,7 +441,7 @@ void md_param_alloc(methoddesc *md)
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->regoff   = stacksize * 8;
 			}
 			stacksize += 2;
 			break;
@@ -530,7 +541,7 @@ void md_param_alloc_native(methoddesc *md)
 					ALIGN_2(stacksize);
 
 					pd->inmemory = true;
-					pd->regoff   = stacksize;
+					pd->regoff   = stacksize * 4;
 				}
 				stacksize += 2;
 			}
@@ -545,7 +556,7 @@ void md_param_alloc_native(methoddesc *md)
 				}
 				else {
 					pd->inmemory = true;
-					pd->regoff   = stacksize;
+					pd->regoff   = stacksize * 4;
 				}
 				stacksize++;
 			}
@@ -570,7 +581,7 @@ void md_param_alloc_native(methoddesc *md)
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->regoff   = stacksize * 4;
 			}
 			stacksize++;
 			break;
@@ -594,7 +605,7 @@ void md_param_alloc_native(methoddesc *md)
 			}
 			else {
 				pd->inmemory = true;
-				pd->regoff   = stacksize;
+				pd->regoff   = stacksize * 4;
 			}
 			stacksize += 2;
 			break;
@@ -630,7 +641,7 @@ void md_param_alloc_native(methoddesc *md)
 
 *******************************************************************************/
 
-void md_return_alloc(jitdata *jd, stackptr stackslot)
+void md_return_alloc(jitdata *jd, stackelement_t *stackslot)
 {
 	methodinfo *m;
 	methoddesc *md;
