@@ -1,9 +1,7 @@
 /* src/vm/jit/reg.h - register allocator header
 
-   Copyright (C) 1996-2005, 2006, 2007 R. Grafl, A. Krall, C. Kruegel,
-   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
-   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
-   J. Wenninger, Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006, 2007, 2008
+   CACAOVM - Verein zur Foerderung der freien virtuellen Maschine CACAO
 
    This file is part of CACAO.
 
@@ -22,8 +20,6 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: reg.h 7713 2007-04-15 21:49:48Z twisti $
-
 */
 
 
@@ -59,7 +55,7 @@ struct varinfo {
 		basicblock *retaddr;
 	} vv;
 #if defined(ENABLE_VERIFIER)
-	typeinfo typeinfo;         /* type info for reference types              */
+	typeinfo_t typeinfo;       /* type info for reference types              */
 #endif
 };
 
