@@ -22,14 +22,14 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: arch.h 7695 2007-04-12 19:49:34Z twisti $
-
 */
 
 
 #ifndef _ARCH_H
 #define _ARCH_H
 
+#define JIT_COMPILER_VIA_SIGNAL
+
 #include "config.h"
 
 
@@ -112,8 +112,7 @@
 
 /* replacement ****************************************************************/
 
-#define REPLACEMENT_PATCH_SIZE           5             /* bytes */
-#define REPLACEMENT_STUB_SIZE            (10+2+10+2+1) /* bytes */
+#define REPLACEMENT_PATCH_SIZE           2             /* bytes */
 
 #endif /* _ARCH_H */
 
