@@ -22,8 +22,6 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: emit.h 7596 2007-03-28 21:05:53Z twisti $
-
 */
 
 
@@ -221,6 +219,7 @@ void emit_movswq_reg_reg(codegendata *cd, s8 reg, s8 dreg);
 void emit_movswq_membase_reg(codegendata *cd, s8 basereg, s8 disp, s8 dreg);
 void emit_movslq_reg_reg(codegendata *cd, s8 reg, s8 dreg);
 void emit_movslq_membase_reg(codegendata *cd, s8 basereg, s8 disp, s8 dreg);
+void emit_movzbq_reg_reg(codegendata *cd, s8 reg, s8 dreg);
 void emit_movzwq_reg_reg(codegendata *cd, s8 reg, s8 dreg);
 void emit_movzwq_membase_reg(codegendata *cd, s8 basereg, s8 disp, s8 dreg);
 void emit_movswq_memindex_reg(codegendata *cd, s8 disp, s8 basereg, s8 indexreg, s8 scale, s8 reg);
@@ -268,7 +267,6 @@ void emit_imul_imm_membase_reg(codegendata *cd, s8 imm, s8 basereg, s8 disp, s8
 void emit_imull_imm_membase_reg(codegendata *cd, s8 imm, s8 basereg, s8 disp, s8 dreg);
 void emit_idiv_reg(codegendata *cd, s8 reg);
 void emit_idivl_reg(codegendata *cd, s8 reg);
-void emit_ret(codegendata *cd);
 void emit_shift_reg(codegendata *cd, s8 opc, s8 reg);
 void emit_shiftl_reg(codegendata *cd, s8 opc, s8 reg);
 void emit_shift_membase(codegendata *cd, s8 opc, s8 basereg, s8 disp);
@@ -295,7 +293,6 @@ void emit_push_reg(codegendata *cd, s8 reg);
 void emit_push_imm(codegendata *cd, s8 imm);
 void emit_pop_reg(codegendata *cd, s8 reg);
 void emit_xchg_reg_reg(codegendata *cd, s8 reg, s8 dreg);
-void emit_nop(codegendata *cd);
 void emit_call_reg(codegendata *cd, s8 reg);
 void emit_call_imm(codegendata *cd, s8 imm);
 void emit_call_mem(codegendata *cd, ptrint mem);
