@@ -22,8 +22,6 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   $Id: properties.h 7246 2007-01-29 18:49:05Z twisti $
-
 */
 
 
@@ -31,25 +29,28 @@
 #define _PROPERTIES_H
 
 #include "config.h"
-#include "vm/types.h"
+
+#include <stdint.h>
 
 #include "vm/global.h"
 
 
 /* function prototypes ********************************************************/
 
-bool  properties_init(void);
-bool  properties_postinit(void);
+void  properties_init(void);
+void  properties_set(void);
 
 void  properties_add(char *key, char *value);
 char *properties_get(char *key);
 
-void  properties_system_add(java_objectheader *p, char *key, char *value);
+void  properties_system_add(java_handle_t *p, char *key, char *value);
 
 #if defined(ENABLE_JAVASE)
-void  properties_system_add_all(java_objectheader *p);
+void  properties_system_add_all(java_handle_t *p);
 #endif
 
+void  properties_dump(void);
+
 #endif /* _PROPERTIES_H */
 
 
